/*
 * board/cmx300/cmx300.c
 *
 * Startup configuration and initialization for the CompuLab CM-X300 module
 *
 * (C) Copyright 2009, 2010
 * Igor Grinberg, CompuLab, grinberg@compulab.co.il
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include <common.h>
#include <asm/arch/hardware.h>
#include <asm/arch/pxa-regs.h>
#include <nand.h>
#include <../drivers/net/dm9000x.h>
#include <da9030.h>
#include "da9030.h"
#include "cmx300_common.h"
#include "mfp.h"
#include "gpio.h"

DECLARE_GLOBAL_DATA_PTR;

#define CMX300_DEBUG	0

#if CMX300_DEBUG
#define pr_dbg printf
#else
#define pr_dbg(fmt, ...)
#endif

/* GPIO_REV should get the value read on GPIOs */
#define GPIO_REV(x)	(((x) == 0x1) ? 190 : (100 + 10 * ((x) + 1)))

#define CONFIG_U_BOOT_MAX_SIZE	0x40000	/* 256KB */
#define SPLASH_SIZE		0xC0000	/* 768KB */

#ifdef CONFIG_USE_IRQ
#define IRQ_STACK_SIZE	(CONFIG_STACKSIZE_IRQ + CONFIG_STACKSIZE_FIQ)
#else
#define IRQ_STACK_SIZE	0
#endif

/* Starting from rev 1.3 led gpio is 76 */
#define GLED_GPIO(revision)	((revision) >= 130 ? 76 : 79)
#define BL_GPIO			19

#define SPLASH_ENV_VAR_NAME	"splashimage"
#define FREQUENCY_ENV_VAR_NAME	"cpufreq"
#define OVERCLCK_ENV_VAR_NAME	"c208_overclock_en"
#define ALTBOOTCMD_ENV_VAR_NAME	"alt_bootcmd"

#define PECR_IN0 (1 << 0)

/* CPU core frequencies, MHz */
enum CPU_frequency {
	CPU_freq_invalid	= 0,
	CPU_freq_104		= 104,
	CPU_freq_208		= 208,
	CPU_freq_416		= 416,
	CPU_freq_624		= 624,
};

union serial_num {
	struct {
		unsigned char mac[2];

		int day:5;
		int month:4;
		int year:15;

		int cpu:2;
		int nor:3;
		int dram:3;
		int nand:3;

		int mg:1;
		int b:1;
		int e:1;
		int r:1;

		int at:2;
		int w:1;
		int bt:1;
	} __attribute__ ((packed)) t;
	struct {
		unsigned int low;
		unsigned int high;
	} d;
};

static char *month[] = {
	"Jan", "Feb", "Mar", "Apr", "May", "Jun",
	"Jul", "Aug", "Sep", "Oct", "Nov", "Dec",
};

/*
 * the grammer is:
 *	- Capital letter for the option name,
 *	- 'd' for the Date,
 *	- semicolon for the MAC address.
 */
int get_ser_option(const char o, const char *s)
{
	int i;
	for (i = 0; s[i] != '\0' || i < 100; i++) {
		if (s[i] == '-') {
			/* deal with Date */
			if (o == 'd' && s[i + 1] >= '0' && s[i + 1] <= '9')
				return i + 1;
			/* deal with MAC */
			if (o == ':' &&
			    s[i + 1] == '0' && s[i + 2] == '0' && s[i + 3] == ':')
				return i + 1;
			/* deal with other options */
			if (o == s[i + 1])
				return i + 2;
		}
	}

	return -1;
}

void get_board_serial(struct tag_serialnr *serialnr)
{
	int i, j;
	union serial_num ser_num;
	u32 cpu = 0, dram = 0, nand = 0;
	char const *serial = getenv(SERIAL_ENV_VAR_NAME);

	memset(&ser_num, 0, sizeof(union serial_num));

	if (!serial)
		return;
	/* get the cpu option: */
	i = get_ser_option('C', serial);
	if (i > 0) {
		cpu = simple_strtoul(&serial[i], NULL, 10);
		/*
		 * 1st lsb bit is cpu speed: 1 - 624Mz, 0 - 208Mz
		 * 2nd lsb bit is M option: PXA310 cpu
		 */
		ser_num.t.cpu = (cpu >> 9) & 0x1;
		ser_num.t.cpu |= serial[i + 3] == 'M' ? 2 : 0;
	}

	/* get the dram option: */
	i = get_ser_option('D', serial);
	if (i > 0) {
		dram = simple_strtoul(&serial[i], NULL, 10);
		/* this gives 0 for 16M, 1 for 32M, 2 for 64 and 4 for 128M */
		ser_num.t.dram = dram >> 5 & 0x7;
	}

	/* get the nand option: */
	i = get_ser_option('N', serial);
	if (i > 0) {
		nand = simple_strtoul(&serial[i], NULL, 10);
		/* 0 for 128M, 1 for 256M, 2 for 512 and 4 for 1024M */
		ser_num.t.nand = nand >> 8 & 0x7;
	}

	/* get the battery option: */
	ser_num.t.b = get_ser_option('B', serial) > 0 ? 1 : 0;

	/* get the ethernet option: */
	ser_num.t.e = get_ser_option('E', serial) > 0 ? 1 : 0;

	/* get the audio/touchscreen option: */
	i = get_ser_option('A', serial);
	if (i > 0)
		ser_num.t.at = serial[i] == 'T' ? 1 << 1 : 1;

	/* get wifi/bluetooth option: */
	ser_num.t.w = get_ser_option('W', serial) > 0 ? 1 : 0;

	/* get RTC option? it should be always on, isn't it? */
	ser_num.t.r = get_ser_option('R', serial) > 0 ? 1 : 0;

	i = get_ser_option('d', serial);
	ser_num.t.day = simple_strtoul(&serial[i], NULL, 10) & 0x1f;
	for (j = 0; j < ARRAY_SIZE(month); j++)
		if (!strncmp(&serial[i + 3], month[j], 3))
			break;
	ser_num.t.month = j & 0xf;
	ser_num.t.year = simple_strtoul(&serial[i + 7], NULL, 10) & 0xfff;

	/* here we pass only the last 2 bytes of MAC address */
	i = get_ser_option(':', serial);
	if (i > 0 && ser_num.t.e) {
		ser_num.t.mac[0] = simple_strtoul(&serial[i + 15], NULL, 16) & 0xff;
		ser_num.t.mac[1] = simple_strtoul(&serial[i + 12], NULL, 16) & 0xff;
	}

	serialnr->low = ser_num.d.low;
	serialnr->high = ser_num.d.high;
}

u32 cmx300_rev;

u32 get_board_rev(void)
{
	return cmx300_rev;
}

u32 get_board_gpio_rev(void)
{
	int rev;
	uint mask = MFPR_DEFAULT | MFP_PULL_SEL | MFP_PULLUP_EN | MFP_AF0;

	/* pullup GPIOs: 108,18,16 in MFPR */
	mfpr_write(MFPR_GPIO108, mask);
	mfpr_write(MFPR_GPIO18, mask);
	mfpr_write(MFPR_GPIO16, mask);

	/* set to input */
	GPDR3 &= ~GPIO_bit(12);		/* GPIO108 */
	GPDR0 &= ~GPIO_bit(18);		/* GPIO18 */
	GPDR0 &= ~GPIO_bit(16);		/* GPIO16 */

	/* wait for the gpios to settle down to read the right values */
	udelay(10);

	/* read them */
	rev = (GPLR3 >> 12) & 0x1;	/* GPIO108 is LSB */
	rev |= (GPLR0 >> 17) & 0x2;
	rev |= (GPLR0 >> 14) & 0x4;	/* GPIO16 is MSB */
	rev = ~rev & 0x7;		/* clear the garbage */

	rev = GPIO_REV(rev);

	/* only from revision 1.3 the revision could be read from gpios */
	if (rev < 130)
		return 0;

	return rev == 140 ? 141 : rev;
}

u32 get_board_env_rev(void)
{
	int i;
	char const *serial = getenv(SERIAL_ENV_VAR_NAME);

	if (!serial)
		return 0;

	/* Revision number format is: -V### */
	i = get_ser_option('V', serial);

	return simple_strtoul(&serial[i], NULL, 10);
}

/* Light up the green led */
void gled_enable(void)
{
	if (cmx300_rev) {
		GPDR2 |= GPIO_bit(GLED_GPIO(cmx300_rev));
		GPCR2 = GPIO_bit(GLED_GPIO(cmx300_rev));
	}
}

void cs_gpio_enable(void)
{
	mfpr_write(MFPR_GPIO1, MFPR_DEFAULT | MFP_AF1);
	mfpr_write(MFPR_GPIO2, MFPR_DEFAULT | MFP_AF1);
}

/* The only valid operating points */
static u32 core_freq_to_accr(int freq)
{
	/* snap down the core frequency to a proper value */
	freq = freq < CPU_freq_104 ?
		CPU_freq_104 : freq < CPU_freq_208 ?
		CPU_freq_104 : freq < CPU_freq_416 ?
		CPU_freq_208 : freq < CPU_freq_624 ?
		CPU_freq_416 : CPU_freq_624;

	switch (freq) {
	case CPU_freq_104:
		/*	XL      XN      DMCFS      HSS      SFLFS     SMCFS */
		return 0x8  |(1 << 8)|(3 << 12)|(0 << 14)|(0 << 18)|(0 << 23);
	case CPU_freq_208:
		return 0x10 |(1 << 8)|(3 << 12)|(0 << 14)|(1 << 18)|(2 << 23);
	case CPU_freq_416:
		return 0x10 |(2 << 8)|(3 << 12)|(1 << 14)|(2 << 18)|(2 << 23);
	case CPU_freq_624:
		return 0x18 |(2 << 8)|(3 << 12)|(2 << 14)|(3 << 18)|(5 << 23);
	default:;
	}

	return 0;
}

static inline uint get_core_freq(void)
{
	u32 acsr = ACSR;

	return 13 * (acsr & ACCR_XL_MASK) * ((acsr & ACCR_XN_MASK) >> 8);
}

static inline uint get_max_core_freq(void)
{
	char const *serial = getenv(SERIAL_ENV_VAR_NAME);
	int i = get_ser_option('C', serial);

	return i < 0 ? CPU_freq_invalid : simple_strtoul(&serial[i], NULL, 10);
}

static void update_core_freq(uint freq)
{
	uint32_t mask = ACCR_XN_MASK | ACCR_XL_MASK;
	uint32_t accr = ACCR;
	uint32_t xclkcfg;
	uint32_t xn_xl = core_freq_to_accr(freq) &
			(ACCR_XN_MASK | ACCR_XL_MASK);

	if (!xn_xl || freq == get_core_freq())
		return;

	accr &= ~(ACCR_XN_MASK | ACCR_XL_MASK | ACCR_FC_MASK);
	accr |= xn_xl;

	/* No clock until core PLL is re-locked */
	accr |= ACCR_FC_MASK;

	/* turbo bit */
	xclkcfg = ((xn_xl & ACCR_XN_MASK) == (2 << 8)) ? 0x3 : 0x2;

	ACCR = accr;
	__asm__("mcr p14, 0, %0, c6, c0, 0\n" : : "r"(xclkcfg));

	while ((ACSR & mask) != (accr & mask))
		cmx300_cpu_relax();
}

static void update_bus_freq(uint corefreq)
{
	uint32_t mask = ACCR_SMC_MASK | ACCR_SRAM_MASK |
			ACCR_HSIO_MASK | ACCR_DDR_MASK;
	uint32_t accr = ACCR;
	uint32_t new_accr = core_freq_to_accr(corefreq) &
				~(ACCR_XN_MASK | ACCR_XL_MASK);

	if (!new_accr || ((ACSR & mask) == (new_accr & mask)))
		return;

	accr &= ~mask;
	accr |= new_accr;

	ACCR = accr;

	while ((ACSR & mask) != (accr & mask))
		cmx300_cpu_relax();
}

static int altboot_button;

/* Check if the alternate boot button is pressed */
static inline void check_alt_boot_button(void)
{
	altboot_button = (PECR & PECR_IN0) ? 1 : 0;
}

static void switch_to_alt_boot(void)
{
	printf("Alternate boot button pressed!\n");

	if (!getenv(ALTBOOTCMD_ENV_VAR_NAME)) {
		printf("%s variable is not set! "
			"Falling back to regular bootcmd.",
			ALTBOOTCMD_ENV_VAR_NAME);
		return;
	}

	setenv("bootdelay", "0");
	setenv("bootcmd", "run " ALTBOOTCMD_ENV_VAR_NAME);
}

extern void pxa310_usb_port3_phy_suspend(void);
extern void pxa310_usb_port3_phy_resume(void);

int board_init(void)
{
	/* MFPR and GPIO are setup in lowlevel_init.S */

	uint cpufreq = 13 * CONFIG_SYS_MONAHANS_RUN_MODE_OSC_RATIO *
				CONFIG_SYS_MONAHANS_TURBO_RUN_MODE_RATIO;

	/* Check if the alternate boot button is pressed */
	check_alt_boot_button();

	/* start at defined frequency */
	update_core_freq(cpufreq);
	update_bus_freq(cpufreq);

	/* arch number of CM-X300 Board */
	gd->bd->bi_arch_number = MACH_TYPE_CM_X300;

	/* address of boot parameters */
	gd->bd->bi_boot_params = 0x80000100;

	/* frame buffer address to U-Boot start + 6M */
	gd->fb_base = TEXT_BASE + 0x600000;

	/*
	 * according to pxa3xx manual vol. II (p.70)
	 * we need to set CSMSADRCFG(2) register
	 */
	CSMSADRCFG = 0x2;

	/* configure the GPIO1 & GPIO2 as CS2 & CS3 accordingly */
	cs_gpio_enable();

	/* configure the GPIO95 as CS for V3020 RTC */
	mfpr_write(MFPR_GPIO95, MFPR_DEFAULT | MFP_PULL_SEL
			| MFP_PULLUP_EN | MFP_SLEEP_DATA | MFP_SLEEP_OE_N);

	/* enable baseboard power (needed for baseboard revision 1.0) */
	GPDR2 |= GPIO_bit(87);
	GPCR2 = GPIO_bit(87);

	/*
	 * revision related initialization
	 * starting from revision 1.3 can be done by reading gpios.
	 */
	cmx300_rev = get_board_gpio_rev();
	gled_enable();

	/* suspend the PHY on USB port 3 to save some power */
	if (cpu_is_pxa310())
		pxa310_usb_port3_phy_suspend();

	return 0;
}

static int adv_pwr_mod;

int is_adv_pwr_mod_enabled(void)
{
	return adv_pwr_mod;
}

static void set_board_freqs(void)
{
	char temp[8];
	char *desired_freq_str = getenv(FREQUENCY_ENV_VAR_NAME);
	uint max_freq = get_max_core_freq();
	uint desired_freq = max_freq;

	printf("CPU:   ");

	if (desired_freq_str)
		desired_freq = simple_strtoul(desired_freq_str, NULL, 10);

	if (max_freq != CPU_freq_invalid && desired_freq > max_freq) {
		if (getenv(OVERCLCK_ENV_VAR_NAME) == NULL) {
			desired_freq = max_freq;
			sprintf(temp, "%d", max_freq);
			setenv(FREQUENCY_ENV_VAR_NAME, temp);
		}
	}

	if (desired_freq != CPU_freq_invalid) {
		update_core_freq(desired_freq);
		update_bus_freq(desired_freq);
	}

	printf("%u MHz\n", get_core_freq());
}

int board_env_ready(void)
{
	struct da9030_ops *ops = da9030_get_ops();
	u32 mask;

	adv_pwr_mod = getenv(ADVANCED_POWER_MODE_ENV_VAR_NAME) ? 1 : 0;

	/* turn on all LDOs for backward compatibility */
	if (!adv_pwr_mod) {
		/* additional preparations to enable LDO10 */
		mask = STARTUP_CTL_LDO10_START | STARTUP_CTL_LDO11_START;
		ops->set_reg(STARTUP_CONTROL, mask);

		mask = RC1_LDO16_EN | RC1_LDO11_EN |
		       RC1_LDO10_EN | RC1_BUCK2_EN;
		ops->set_reg(REG_CONTROL_1_17, mask);

		mask = RC3_LDO7_EN | RC3_LDO6_EN | RC3_LDO5_EN |
		       RC3_LDO3_EN | RC3_BUCK_EN;

		/* In PXA310 LDO4 is connected to ULPI PHY */
		if (cpu_is_pxa310())
			mask |= RC3_LDO4_EN;

		ops->set_reg(REG_CONTROL_1_97, mask);

		mask = RC4_LDO12_EN | RC4_LDO11_EN | RC4_LDO10_EN;
		ops->set_reg(REG_CONTROL_2_98, mask);
	}

	set_board_freqs();

	if (!cmx300_rev) {
		/*
		 * revision related initialization
		 * performed in board_init() starting only from revision 1.3
		 */
		cmx300_rev = get_board_env_rev();
		gled_enable();
	}

	/* In compatibility mode: enable USB Charge Pump for rev. >=1.3 */
	if (!adv_pwr_mod && get_board_rev() >= 130) {
		ops->set_reg(MISC_CONTROLB, MISCB_SESSION_VALID_ENABLE);
		mask = USB_PUMP_EN_USBVEP | USB_PUMP_EN_USBVE | USB_PUMP_USBVE;
		ops->set_reg(USBPUMP, mask);
		mdelay(100);

		/* enable the PHY on USB port 3 */
		if (cpu_is_pxa310())
			pxa310_usb_port3_phy_resume();
	}

	set_user_gpios();

	return 0;
}

int misc_init_r(void)
{
	setenv("stdout", "serial");
	setenv("stderr", "serial");

#ifdef CONFIG_SILENT_CONSOLE
	if (getenv("silent") != NULL)
		gd->flags |= GD_FLG_SILENT;
	else
		gd->flags &= ~GD_FLG_SILENT;
#endif

	return 0;
}

int board_late_init(void)
{
	char fb_base_str[11];
	uint rev_major = cmx300_rev / 100;
	uint rev_minor = cmx300_rev - (100 * rev_major);

	/* set frame buffer base address for WinCE */
	sprintf(fb_base_str, "%lx", gd->fb_base);
	setenv("fbaddr", fb_base_str);

	/* if there is splash screen, enable the backlight */
	if (getenv(SPLASH_ENV_VAR_NAME)) {
		GPDR0 |= GPIO_bit(BL_GPIO);
		GPSR0 = GPIO_bit(BL_GPIO);
	}

	if (ctrlc()) {
		gd->flags &= ~GD_FLG_SILENT;
		setenv("bootdelay", "-1");
	}

	if (altboot_button)
		switch_to_alt_boot();

	if ((rev_minor / 10) * 10 == rev_minor)
		rev_minor = rev_minor / 10;

	printf("PCB:   %u.%u\n", rev_major, rev_minor);

	return 0;
}

/*
 * Here we assume that OBM has configured the DRAM correctly,
 * so we just read the configuration.
 */
static inline uint get_sdram2_size(void)
{
	return MDCNFG & MDCNFG_DCSE1 ? PHYS_SDRAM_2_SIZE : 0;
}

/* CM-X300 modules can come either with 64MB or 128MB RAM */
int dram_init(void)
{
	gd->bd->bi_dram[0].start = PHYS_SDRAM_1;
	gd->bd->bi_dram[0].size = PHYS_SDRAM_1_SIZE;
	gd->bd->bi_dram[1].start = PHYS_SDRAM_2;
	gd->bd->bi_dram[1].size = get_sdram2_size();

	return 0;
}

static int get_splash_dram_bank(u32 addr)
{
	if (addr >= PHYS_SDRAM_1 &&
	    addr <= PHYS_SDRAM_1 + PHYS_SDRAM_1_SIZE - SPLASH_SIZE)
		return 1;

	if (addr >= PHYS_SDRAM_2 &&
	    addr <= PHYS_SDRAM_2 + PHYS_SDRAM_2_SIZE - SPLASH_SIZE)
		return 2;

	return -1;
}

static int is_overlapping(u32 low1, u32 high1, u32 low2, u32 high2)
{
	if (low1 >= low2 && low1 <= high2)
		return 1;

	if (high1 >= low2 && high1 <= high2)
		return 1;

	if (low1 <= low2 && high1 >= high2)
		return 1;

	return 0;
}

static int splash_check_overlap(u32 addr)
{
	int splash_dram_bank = get_splash_dram_bank(addr);
	int u_boot_lower_bound = TEXT_BASE -
				CONFIG_SYS_MALLOC_LEN -
				CONFIG_SYS_GBL_DATA_SIZE -
				IRQ_STACK_SIZE -
				0x100000;	/* assume stack size > 1MB */

	/* check RAM bounds */
	if ((splash_dram_bank != 1 && splash_dram_bank != 2) ||
	    (splash_dram_bank == 2 && get_sdram2_size() == 0))
		return 1;
	else if (splash_dram_bank == 2)	/* DRAM bank 2 is not used by U-Boot */
		return 0;

	/* check frame buffer bounds, frame buffer shouldn't be more than 1M */
	if (is_overlapping(addr, addr + SPLASH_SIZE,
			   gd->fb_base, gd->fb_base + 0x100000))
		return 1;

	/* check U-Boot and stack bounds, assume that 1M is enough for stack */
	if (is_overlapping(addr, addr + SPLASH_SIZE,
			   u_boot_lower_bound,
			   TEXT_BASE + CONFIG_U_BOOT_MAX_SIZE))
		return 1;

	return 0;
}

int board_splash_screen_prepare(void)
{
	size_t size = SPLASH_SIZE;
	char *splash_str = getenv(SPLASH_ENV_VAR_NAME);
	u32 splash_addr;

	if (!splash_str)
		return 1;

	splash_addr = simple_strtoul(splash_str, NULL, 16);

	if (splash_check_overlap(splash_addr)) {
		printf("Cannot relocate splash screen image! "
			"%s variable is invalid!\n", SPLASH_ENV_VAR_NAME);
		return 1;
	}

	return nand_read_skip_bad(&nand_info[0], 0xC0000, &size,
						(u_char *)splash_addr);
}

int board_eth_init(bd_t *bis)
{
	int ret, count = 0;
	u32 val;
	char *ethaddr = getenv("ethaddr"), *eth1addr = getenv("eth1addr");
	char *extdm9000 = getenv("extdm9000");
	struct da9030_ops *ops = da9030_get_ops();

	ret = dm9000_eth_initialize(DM9000_IO, DM9000_DATA);
	if (ret == 0 && ethaddr)
		dm9000_set_mac_addr(count, gd->bd->bi_enetaddr);

	if (ret == 0)
		count++;
	else {
		/* The Ethernet doesn't exist, disable LDO3 */
		ops->get_reg(REG_CONTROL_1_97, &val);
		val &= ~RC3_LDO3_EN;
		ops->set_reg(REG_CONTROL_1_97, val);
	}

#if defined(CONFIG_NET_MULTI)	&& defined(CONFIG_HAS_ETH1) \
 && defined(EXT_DM9000_IO)	&& defined(EXT_DM9000_DATA)
	if (extdm9000 && eth1addr) {
		/* Configure the CS3 for use with external DM9000 */
		CSADRCFG3 = 0x12c919;

		ret = dm9000_eth_initialize(EXT_DM9000_IO, EXT_DM9000_DATA);
		if (ret == 0 && (eth1addr || (count == 0 && ethaddr)))
			dm9000_set_mac_addr(count, gd->bd->bi_enet1addr);

		if (ret == 0)
			count++;
	}
#endif

	return count ? count : -1;
}

int mmc_board_init(void)
{
	mfpr_write(MFPR_GPIO3, MFPR_DEFAULT | MFP_AF4);
	mfpr_write(MFPR_GPIO4, MFPR_DEFAULT | MFP_AF4);
	mfpr_write(MFPR_GPIO5, MFPR_DEFAULT | MFP_AF4);
	mfpr_write(MFPR_GPIO6, MFPR_DEFAULT | MFP_AF4);
	mfpr_write(MFPR_GPIO7, MFPR_DEFAULT | MFP_AF4);
	mfpr_write(MFPR_GPIO8, MFPR_DEFAULT | MFP_AF4);

	return 0;
}

