/*
 * board/cmx300/lcd.c
 *
 * TOPPOLY LCD reset and startup for the CompuLab CM-X300 module
 *
 * (C) Copyright 2010
 * Igor Grinberg, CompuLab, grinberg@compulab.co.il
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include <common.h>
#include <pca953x.h>
#include "cmx300_common.h"

#define LCD_DEBUG	0

#define LCD_I2C_GPIO_ADDR	0x25

#define GPIO_LCD_DIN		8	/* aux_gpio3_0 */
#define GPIO_LCD_DOUT		9	/* aux_gpio3_1 */
#define GPIO_LCD_SCL		10	/* aux_gpio3_2 */
#define GPIO_LCD_CS		11	/* aux_gpio3_3 */
#define GPIO_LCD_RESET		12	/* aux_gpio3_4 */

#define TD035STEE1_LCD_ID	0x2008

#if LCD_DEBUG
#define pr_debug printf
#else
#define pr_debug(fmt, ...)
#endif

/* lcd_gpio_set_value:
 * @bit   - one of LCD_X_BITs
 * @value - 0 for low or 1 for high
 * Note: you can only set value for pins configured as output!
 * Trying to set value for pins not configured as output will have no effect!
 */
static inline int lcd_gpio_set_value(uint bit, uint value)
{
	return pca953x_set_val(LCD_I2C_GPIO_ADDR, 1 << bit, value << bit);
}

/* lcd_gpio_get_value:
 * @bit   - one of LCD_X_BITs
 */
static inline int lcd_gpio_get_value(uint bit)
{
	return (pca953x_get_val(LCD_I2C_GPIO_ADDR) >> bit) & 0x1;
}

/* lcd_gpio_direction_output:
 * @bit   - one of LCD_X_BITs
 * @value - 0 for low or 1 for high
 */
static int lcd_gpio_direction_output(uint bit, uint value)
{
	int ret = pca953x_set_dir(LCD_I2C_GPIO_ADDR,
				  1 << bit, PCA953X_DIR_OUT << bit);
	if (ret)
		return ret;

	return lcd_gpio_set_value(bit, value);
}

/* lcd_gpio_direction_input:
 * @bit   - one of LCD_X_BITs
 */
static inline int lcd_gpio_direction_input(uint bit)
{
	return pca953x_set_dir(LCD_I2C_GPIO_ADDR,
				1 << bit, PCA953X_DIR_IN << bit);
}

static void lcd_write_bits(int data, int nbits, int clk_active_low)
{
	int i = 0;

	lcd_gpio_set_value(GPIO_LCD_SCL, 0);
	mdelay(1);

	for (i = 0; i < nbits; ++i) {
		lcd_gpio_set_value(GPIO_LCD_DIN,
				   (data & (1 << (nbits - 1 - i))) ? 1 : 0);
		mdelay(1);

		lcd_gpio_set_value(GPIO_LCD_SCL, 1);
		mdelay(1);

		if (clk_active_low && (nbits - 1 == i))
			lcd_gpio_set_value(GPIO_LCD_SCL, 1);
		else
			lcd_gpio_set_value(GPIO_LCD_SCL, 0);

		mdelay(1);
	}

	if (clk_active_low) {
		lcd_gpio_set_value(GPIO_LCD_SCL, 1);
		mdelay(1);
	}
}

static void lcd_write_read_bits(int in_data, int nbits,
				int *out_data, int clk_active_low)
{
	int i, temp = 0;
	*out_data = 0;

	lcd_gpio_set_value(GPIO_LCD_CS, 0);
	mdelay(1);

	lcd_gpio_set_value(GPIO_LCD_SCL, 0);
	mdelay(1);

	for (i = 0 ; i < nbits; ++i) {
		lcd_gpio_set_value(GPIO_LCD_DIN,
				   (in_data & (1 << (nbits - 1 - i))) ? 1 : 0);
		mdelay(1);

		lcd_gpio_set_value(GPIO_LCD_SCL, 1);
		mdelay(1);

		temp = lcd_gpio_get_value(GPIO_LCD_DOUT);
		*out_data |= (((temp == 1) ? 1 : 0) << (nbits - 1 - i));
		mdelay(1);

		lcd_gpio_set_value(GPIO_LCD_SCL, 0);
		mdelay(1);
	}
	if (clk_active_low) {
		lcd_gpio_set_value(GPIO_LCD_SCL, 1);
		mdelay(1);
	}

	lcd_gpio_set_value(GPIO_LCD_CS, 1);
	mdelay(1);
}

static void send_ssp_9bit(u8 data)
{
	lcd_gpio_set_value(GPIO_LCD_CS, 0);
	mdelay(1);

	lcd_write_bits(data, 9, 1);
	mdelay(1);

	lcd_gpio_set_value(GPIO_LCD_CS, 1);
}

/* static void send_ssp_18bit(u8 x, u8 x1) */
static void send_ssp_18bit(int data)
{
	lcd_gpio_set_value(GPIO_LCD_CS, 0);
	mdelay(1);

	lcd_write_bits(data, 18, 1);
	mdelay(1);

	lcd_gpio_set_value(GPIO_LCD_CS, 1);
}

static void cm_x300_init_lcd_gpios(void)
{
	lcd_gpio_direction_output(GPIO_LCD_DIN, 0);
	lcd_gpio_direction_input(GPIO_LCD_DOUT);
	lcd_gpio_direction_output(GPIO_LCD_SCL, 0);
	lcd_gpio_direction_output(GPIO_LCD_CS, 1);
	lcd_gpio_direction_output(GPIO_LCD_RESET, 1);
}

int toppoly_lcd_init(void)
{
	int i;
	int lcd_data = 0;

	/* Initialize GPIOs and reset the LCD controller */
	cm_x300_init_lcd_gpios();
	mdelay(10);

	for (i = 0; i < 3; ++i) {
		send_ssp_9bit(0x0);
		mdelay(1);
	}

	lcd_gpio_set_value(GPIO_LCD_SCL, 1);
	mdelay(1);

	/* Address 0xBC and data 0x80 */
	send_ssp_18bit(0x17980);
	mdelay(1);

	/* Address 0xBF and data 0x10 */
	send_ssp_18bit(0x17f10);
	mdelay(1);

	/* Send msleep out command */
	send_ssp_9bit(0x11);
	mdelay(50);

	/* Send DISPLAY_ON command */
	send_ssp_9bit(0x29);
	mdelay(5);

	/* Send Display power mode command [shifted by 0x4] */
	lcd_write_read_bits(0x02000000, 32, &lcd_data, 1);

	/* Checking match of LCD ID */
	if ((lcd_data & 0xFFFF) != TD035STEE1_LCD_ID) {
		pr_debug("%s:   Failed to initialize the LCD!", "LCD");
		pr_debug(" ID does not match: %x\n", lcd_data);
		return 1;
	}
	pr_debug("%s:   OK\n", "LCD");

	mdelay(100);

	return 0;
}

